// lookup = (name) => returns <config>

class Config {
  constructor(config) {
    this.config = {
        "operators": [],
        "bridges": [],
        "hierarchy": [],
        "priorities": [],
        "version": "3",
        "debug": false,
        "associations": {
            'negative': [],
            'positive': []
           },
        "objects": {},
        "description": "",
        "words": {},
        "floaters": [],
        "implicits": [],
        "flatten": [],
        "utterances": [],
        "contexts": [],
        "generators": [],
        "semantics": [],

        "debug": false,
        "expected_generated": [],
        "expected_results": [],
        "skipSemantics": false
    }
    this.initializerFN = () => {}
    this.wasInitialized = false

    if (config) {
      this.checkBridges(config.bridges)
      this.add(config);
    }
  }

  initializer( fn ) {
    this.initializerFN = fn
  }

  initialize() {
    if (this.wasInitialized) {
      return; // only one run per customer
    }
    this.initializerFN(this)
    this.wasInitialized = true
  }

  checkBridges(bridges) {
    const keys = {}
    bridges.forEach( (bridge) => {
      const key = JSON.stringify({ id: bridge.id, level: bridge.level })
      if (keys[key]) {
        throw `The operator ${bridge.id}/${bridge.level} appears more than once in the bridges.`
      } else {
        keys[key] = true
      }
    })
  };

  get(property) {
    return this.config[property]
  }

  set(property, value) {
    this.config[property] = value;
  }

  add(more) {
    if (more instanceof Config) {
      more.initialize();
      more = more.config
    }
    for (const key of Object.keys(more)) {
      const value = more[key]
      if (key == 'words') {
        const config_words = this.config['words']
        const more_words = more['words']
        for (const word of Object.keys(more_words)) {
          if (!config_words[word]) {
            config_words[word] = []
          }
          config_words[word] = config_words[word].concat(more_words[word])
        }
      } else if (key == 'debug') {
        this.config[key] = this.config[key] || more[key]
      } else if (key == 'description') {
        this.config[key] += more[key]
      } else if (key == 'objects') {
        Object.assign(this.config[key], more[key])
      } else if (key == 'associations') {
        const config_associations = this.config['associations']
        const more_associations = more['associations']
        if (more_associations['positive']) {
          config_associations['positive'] = config_associations['positive'].concat(more_associations['positive'])
        }
        if (more_associations['negative']) {
          config_associations['negative'] = config_associations['negative'].concat(more_associations['negative'])
        }
      } else if (Array.isArray(value)) {
        this.config[key] = this.config[key].concat(more[key])
      } else {
        if (!(key in this.config)) {
          throw `Unexpected property in config ${key}`
        }
        this.config[key] = more[key]
      }
    }
    return this;
  }
}

module.exports = Config;
